""" Testing utilities that allow for easier assertions on collections.
Do you love tests? Neither do we.
"""
# TODO: This module due is for a refactor.
from typing import Any
import numbers
import numpy as np
import torch
import pandas as pd
from scipy.sparse import csc_matrix
from scipy.sparse import csr_matrix
from scipy.sparse import dok_matrix
from scipy.sparse import lil_matrix


def coerced(o: Any):
    if isinstance(o, (csc_matrix, csr_matrix, dok_matrix, lil_matrix)):
        o = o.toarray()

    to_list_collections = tuple([np.ndarray, torch.Tensor, pd.core.series.Series])
    if isinstance(o, (list, tuple)):
        return o
    elif isinstance(o, to_list_collections):
        return o.tolist()
    elif isinstance(o, pd.core.frame.DataFrame):
        return o.values.tolist()
    elif isinstance(o, dict):
        li = [np.nan] * len(o)
        for k, v in o.items():
            li[k] = v
        return li
    else:
        raise ValueError(f"Object {o} of {type(o)} is not a list, tuple nor array.")


def is_close(
    a: numbers.Number, b: numbers.Number, rel_tol: float = 1e-09, abs_tol: float = 0.0
):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def ctr_eq(c1: Any, c2: Any):
    if isinstance(c1, torch.Tensor) and c1.shape == torch.Size([]):
        c1 = c1.item()
    if isinstance(c2, torch.Tensor) and c2.shape == torch.Size([]):
        c2 = c2.item()

    if isinstance(c1, numbers.Number) and isinstance(c2, numbers.Number):
        return is_close(c1, c2)

    c1 = coerced(c1)
    c2 = coerced(c2)

    return all([ctr_eq(c1[i], c2[i]) for i in range(max(len(c1), len(c2)))])



""" Basic tests for slicer.
An unholy balance of use cases and test coverage.
"""

import pytest

from .slicer import AtomicSlicer

from . import Slicer as S
from . import Alias as A
from . import Obj as O

import pandas as pd
import numpy as np
import torch
from scipy.sparse import csc_matrix
from scipy.sparse import csr_matrix
from scipy.sparse import dok_matrix
from scipy.sparse import lil_matrix


from .utils_testing import ctr_eq


def test_slicer_ragged_numpy():
    values = np.array([
        np.array([0, 1]),
        np.array([2, 3, 4])
    ], dtype=object)
    data = np.array([
        np.array([5, 6, 7]),
    ])

    slicer = S(values=values, data=data)
    sliced = slicer[0, 1]

    assert ctr_eq(sliced.data, data[0][1])
    assert ctr_eq(sliced.values, values[0][1])


def test_slicer_basic():
    data = [[1, 2], [3, 4]]
    values = [[5, 6], [7, 8]]
    identifiers = ["id1", "id1"]
    instance_names = ["r1", "r2"]
    feature_names = ["f1", "f2"]
    full_name = "A"

    slicer = S(
        data=data,
        values=values,
        identifiers=A(identifiers, 0),
        instance_names=A(instance_names, 0),
        feature_names=A(feature_names, 1),
        full_name=full_name,
    )

    colon_actual = slicer[:, 1]
    assert colon_actual.data == [2, 4]
    assert colon_actual.instance_names == ["r1", "r2"]
    assert colon_actual.feature_names == "f2"
    assert colon_actual.values == [6, 8]

    ellipses_actual = slicer[..., 1]
    assert ellipses_actual.data == [2, 4]
    assert ellipses_actual.instance_names == ["r1", "r2"]
    assert ellipses_actual.feature_names == "f2"
    assert ellipses_actual.values == [6, 8]

    array_index_actual = slicer[[0, 1], 1]
    assert array_index_actual.data == [2, 4]
    assert array_index_actual.feature_names == "f2"
    assert array_index_actual.instance_names == ["r1", "r2"]
    assert array_index_actual.values == [6, 8]

    alias_actual = slicer["r1", "f2"]
    assert alias_actual.data == 2
    assert alias_actual.feature_names == "f2"
    assert alias_actual.instance_names == "r1"
    assert alias_actual.values == 6

    alias_actual = slicer["id1", "f2"]
    assert alias_actual.data == [2, 4]
    assert alias_actual.feature_names == "f2"
    assert alias_actual.instance_names == ["r1", "r2"]
    assert alias_actual.values == [6, 8]

    chained_actual = slicer[:][:, 1]
    assert chained_actual.data == [2, 4]
    assert chained_actual.feature_names == "f2"
    assert chained_actual.instance_names == ["r1", "r2"]
    assert chained_actual.values == [6, 8]

    alias_actual = slicer["id1"][:, "f2"]
    assert alias_actual.data == [2, 4]
    assert alias_actual.feature_names == "f2"
    assert alias_actual.instance_names == ["r1", "r2"]
    assert alias_actual.values == [6, 8]

    alias_actual = slicer["r1"]
    alias_actual = alias_actual["f2"]
    assert alias_actual.data == 2
    assert alias_actual.feature_names == "f2"
    assert alias_actual.instance_names == "r1"
    assert alias_actual.values == 6


def test_slicer_unnamed():
    a = [1, 2, 3]
    b = [4, 5, 6]

    slicer = S(a, b)
    actual_a, actual_b = slicer[1].o
    assert actual_a == 2
    assert actual_b == 5

    df1 = pd.DataFrame([[1, 2], [3, 4]])
    df2 = pd.DataFrame([[5, 6], [7, 8]])
    slicer = S(df1, df2)
    actual_1, actual_2 = slicer[:, 0].o

    assert ctr_eq(actual_1.values, [1, 3])
    assert ctr_eq(actual_2.values, [5, 7])


def test_slicer_crud():
    data = [[1, 2], [3, 4]]
    values = [[5, 6], [7, 8]]
    extra = [[9, 10], [11, 12]]
    overridden = [[13, 14], [15, 16]]

    slicer = S(data=data, values=values)
    slicer.extra = extra  # Create
    slicer.data = overridden  # Update
    del slicer.values  # Delete

    sliced = slicer[0, 1]  # Read
    assert sliced.data == 14
    with pytest.raises(Exception):
        _ = sliced.values

    assert sliced.extra == 10

    del slicer.o
    assert slicer.o == []


def test_slicer_default_alias():

    df = pd.DataFrame([[1, 2], [3, 4]], columns=["A", "B"])
    slicer = S(df)
    assert getattr(slicer, "index", None)
    assert getattr(slicer, "columns", None)
    actual = slicer[:, "A"].o
    assert ctr_eq(actual, [1, 3])


def test_slicer_anon_dict():
    di = {"a": [1, 2, 3], "b": [4, 5, 6]}
    slicer = S(di)

    result = slicer["a", 1].o
    assert result == 2


def test_slicer_3d():
    data_2d = [[1, 2], [3, 4], [5, 6]]
    values_3d = [
        [[1, 2, 3], [4, 5, 6]],
        [[7, 8, 9], [10, 11, 12]],
        [[13, 14, 15], [16, 17, 18]],
    ]
    names = ["a", "b", "c"]

    slicer = S(data=data_2d, values=values_3d, names=A(names, 2))
    actual = slicer[..., 1]
    assert ctr_eq(actual.data, data_2d)
    assert actual.names == "b"

    actual = slicer[0, :, 1]
    assert ctr_eq(actual.data, data_2d[0])
    assert actual.names == "b"

    actual = slicer[0, :][:, 1]
    assert ctr_eq(actual.data, data_2d[0])
    assert actual.names == "b"


def test_untracked():
    data = [1, 2, 3, 4]
    primitive = 1
    collection = [[8, 9]]
    slicer = S(data=data, primitive=O(primitive, None), collection=O(collection, None))
    actual = slicer[:2]
    assert actual.data == data[:2]
    assert actual.primitive == primitive
    assert ctr_eq(actual.collection, collection)


def test_partial_untracked():
    s = S(a=np.zeros((4, 5, 6)), b=O(np.ones((4, 2, 2)), [0]))
    assert s[:, :, 1].b.shape == (4, 2, 2)


def test_numpy_subkeys():
    data = [1, 2, 3, 4]
    slicer = S(data=data)

    subkey = np.int64(1)
    assert slicer[subkey].data == 2

    subkey = np.array([0, 1])
    assert ctr_eq(slicer[subkey].data, [1, 2])

    subkey = np.array([[0, 1], [3, 4]])
    with pytest.raises(ValueError):
        _ = slicer[subkey]


def test_repr_smoke():
    slicer = S([1, 2], ["a", "b"], named=[3, 4])
    print(slicer)

    atomic = AtomicSlicer([1, 2, 3, 4])
    print(atomic)


def test_slicer_simple_di():
    di = {"A": [1, 2], "B": [3, 4], "C": [5, 6]}
    slicer = S(di)
    actual = slicer["B", 0]
    actual = actual.o
    assert ctr_eq(actual, 3)

    nested_di = {"X": di, "Y": di}
    actual = S(nested_di)["X", "B", 0].o
    assert ctr_eq(actual, 3)


def test_slicer_sparse():
    array = np.array([[1, 0, 4], [0, 0, 5], [2, 3, 6]])
    csc_array = csc_matrix(array)
    csr_array = csr_matrix(array)
    dok_array = dok_matrix(array)
    lil_array = lil_matrix(array)

    candidates = [csc_array, csr_array, dok_array, lil_array]
    for candidate in candidates:
        print("testing:", type(candidate))
        slicer = S(candidate)
        actual = slicer[0, 0]
        assert ctr_eq(actual.o, 1)
        actual = slicer[1, 1]
        assert ctr_eq(actual.o, 0)

        actual = slicer[0]
        expected = np.array([1, 0, 4])
        assert ctr_eq(actual.o, expected)

        actual = slicer[:, 1]
        expected = np.array([0, 0, 3])
        assert ctr_eq(actual.o, expected)

        actual = slicer[:, :]
        expected = np.array([[1, 0, 4], [0, 0, 5], [2, 3, 6]])
        assert ctr_eq(actual.o, expected)

        actual = slicer[0, :]
        expected = np.array([1, 0, 4])
        assert ctr_eq(actual.o, expected)


def test_slicer_torch():
    import torch

    data = torch.tensor([[1, 2], [3, 4]])
    values = torch.tensor([[5, 6], [7, 8]])
    alias = ["f1", "f2"]

    slicer = S(data=data, values=values, alias=A(alias, 1))
    sliced = slicer[0, "f2"]
    assert sliced.data == 2
    assert sliced.values == 6


def test_slicer_pandas():
    di = {"A": [1, 2], "B": [3, 4], "C": [5, 6]}
    df = pd.DataFrame(di)

    slicer = S(df)
    assert slicer[0, "A"].o == 1
    assert ctr_eq(slicer[:, "A"].o, [1, 2])
    assert ctr_eq(slicer[0, :].o, [1, 3, 5])

    df = pd.DataFrame(di, index=["X", "Y"])
    slicer = S(df)
    assert slicer["X", "A"].o == 1
    assert slicer[0, "A"].o == 1
    assert slicer[0, 0].o == 1
    slicer = S(df["A"])
    assert slicer["X"].o == 1
    assert slicer[0].o == 1
    assert ctr_eq(slicer[:].o, [1, 2])


def test_handle_newaxis_ellipses():
    from .slicer_internal import _handle_newaxis_ellipses

    index_tup = (1,)
    max_dim = 3

    expanded_index_tup = _handle_newaxis_ellipses(index_tup, max_dim)
    assert expanded_index_tup == (1, slice(None), slice(None))
    
    
def test_tracked_dim_arg_smoke():
    li = ['A', 'B']
    _ = A(li, dim=0)
    _ = A(li, dim=[0])
    _ = A(li, dim=(0,))

    # Aliases must have a single dim
    with pytest.raises(Exception):
        _ = A(li, dim=None)

    with pytest.raises(Exception):
        _ = A(li, dim=[0,1])

    _ = O(li, dim=0)
    _ = O(li, dim=[0])
    _ = O(li, dim=(0,))

    assert True


def test_operations_1d():
    elements = [1, 2, 3, 4]
    li = elements
    tup = tuple(elements)
    di = {i: x for i, x in enumerate(elements)}
    series = pd.Series(elements)
    array = np.array(elements)
    torch_array = torch.tensor(elements)
    containers = [li, tup, array, torch_array, di, series]
    for ctr in containers:
        print("testing:", type(ctr))
        slicer = AtomicSlicer(ctr)

        assert ctr_eq(slicer[0], elements[0])

        # Array
        assert ctr_eq(slicer[[0, 1, 2, 3]], elements)
        assert ctr_eq(slicer[[0, 1, 2]], elements[:-1])

        # All
        assert ctr_eq(slicer[:], elements[:])
        assert ctr_eq(slicer[tuple()], elements)

        # Ranged slicing
        if not isinstance(ctr, dict):  # Do not test on dictionaries.
            assert ctr_eq(slicer[-1], elements[-1])
            assert ctr_eq(slicer[0:3:2], elements[0:3:2])


def test_operations_2d():
    elements = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    li = elements
    df = pd.DataFrame(elements, columns=["A", "B", "C"])

    sparse_csc = csc_matrix(elements)
    sparse_csr = csr_matrix(elements)
    sparse_dok = dok_matrix(elements)
    sparse_lil = lil_matrix(elements)

    containers = [li, df, sparse_csc, sparse_csr, sparse_dok, sparse_lil]
    for ctr in containers:
        print("testing:", type(ctr))
        slicer = AtomicSlicer(ctr)

        assert ctr_eq(slicer[0], elements[0])

        # Ranged slicing
        if not isinstance(ctr, dict):
            assert ctr_eq(slicer[-1], elements[-1])
            assert ctr_eq(slicer[0, 0:3:2], elements[0][0:3:2])

        # Array
        assert ctr_eq(slicer[[0, 1, 2], :], elements)

        # All
        assert ctr_eq(slicer[:], elements)
        assert ctr_eq(slicer[tuple()], elements)

        assert ctr_eq(slicer[:, 0], [elements[i][0] for i, _ in enumerate(elements)])
        assert ctr_eq(slicer[[0, 1], 0], [elements[i][0] for i in [0, 1]])
        assert ctr_eq(slicer[[0, 1], 1], [elements[i][1] for i in [0, 1]])
        assert ctr_eq(slicer[0, :], elements[0])
        assert ctr_eq(slicer[0, 1], elements[0][1])

        assert ctr_eq(slicer[..., 0], [elements[i][0] for i, _ in enumerate(elements)])


def test_operations_3d():
    # 3-dimensional fixed dimension case
    elements = [
        [[1, 2, 3], [4, 5, 6]],
        [[7, 8, 9], [10, 11, 12]],
        [[13, 14, 15], [16, 17, 18]],
    ]
    tuple_elements = (
        ((1, 2, 3), (4, 5, 6)),
        ((7, 8, 9), (10, 11, 12)),
        ((13, 14, 15), (16, 17, 18)),
    )
    torch_array = torch.tensor(elements)
    multi_array = np.array(elements)
    list_of_lists = elements
    tuples_of_tuples = tuple_elements
    list_of_multi_arrays = [
        np.array(elements[0]),
        np.array(elements[1]),
        np.array(elements[2]),
    ]
    di_of_multi_arrays = {
        0: np.array(elements[0]),
        1: np.array(elements[1]),
        2: np.array(elements[2]),
    }

    containers = [
        torch_array,
        multi_array,
        tuples_of_tuples,
        list_of_lists,
        list_of_multi_arrays,
        di_of_multi_arrays,
    ]
    for ctr in containers:
        print("testing:", type(ctr))
        slicer = AtomicSlicer(ctr)

        assert ctr_eq(slicer[0], elements[0])

        # Ranged slicing
        if not isinstance(ctr, dict):
            assert ctr_eq(slicer[-1], elements[-1])
            assert ctr_eq(slicer[0, 0:3:2], elements[0][0:3:2])

        # Array
        assert ctr_eq(slicer[[0, 1, 2], :], elements)

        # All
        assert ctr_eq(slicer[:], elements)
        assert ctr_eq(slicer[tuple()], elements)

        assert ctr_eq(slicer[:, 0], [elements[i][0] for i, _ in enumerate(elements)])
        assert ctr_eq(slicer[[0, 1], 0], [elements[i][0] for i in [0, 1]])
        assert ctr_eq(slicer[[0, 1], 1], [elements[i][1] for i in [0, 1]])
        assert ctr_eq(slicer[0, :], elements[0])
        assert ctr_eq(slicer[0, 1], elements[0][1])

        rows = []
        for i, _ in enumerate(elements):
            cols = []
            for j, _ in enumerate(elements[i]):
                cols.append(elements[i][j][1])
            rows.append(cols)
        assert ctr_eq(slicer[..., 1], rows)
        assert ctr_eq(
            slicer[0, ..., 1], [elements[0][i][1] for i in range(len(elements[0]))]
        )

def test_attribute_assignment():
    data = [[1, 2], [3, 4]]
    values = [[5, 6], [7, 8]]
    identifiers = ["id1", "id1"]
    instance_names = ["r1", "r2"]
    feature_names = ["f1", "f2"]
    full_name = "A"

    exp = S(
        data=data,
        values=values,
        identifiers=A(identifiers, 0),
        instance_names=A(instance_names, 0),
        feature_names=A(feature_names, 1),
        full_name=full_name,
    )

    exp.feature_names = ['f3', 'f4']

    assert exp.feature_names == ['f3', 'f4']
    assert exp[:, 0].feature_names == 'f3'
    
    with pytest.raises(Exception):
        _ = exp[:, 'f1'] # f1 should no longer exist as valid alias

    exp.feature_names = A(['f5', 'f6'], dim=0)

    assert exp.feature_names == ['f5', 'f6']
    assert exp[1, :].feature_names == 'f6' # feature_names now tracks dim 0



""" Lower level layer for slicer.
Mom's spaghetti.
"""
# TODO: Consider boolean array indexing.

from typing import Any, AnyStr, Union, List, Tuple
from abc import abstractmethod
import numbers


class AtomicSlicer:
    """ Wrapping object that will unify slicing across data structures.

    What we support:
        Basic indexing (return references):
        - (start:stop:step) slicing
        - support ellipses
        Advanced indexing (return references):
        - integer array indexing

    Numpy Reference:
        Basic indexing (return views):
        - (start:stop:step) slicing
        - support ellipses and newaxis (alias for None)
        Advanced indexing (return copy):
        - integer array indexing, i.e. X[[1,2], [3,4]]
        - boolean array indexing
        - mixed array indexing (has integer array, ellipses, newaxis in same slice)
    """

    def __init__(self, o: Any, max_dim: Union[None, int, AnyStr] = "auto"):
        """ Provides a consistent slicing API to the object provided.

        Args:
            o: Object to enable consistent slicing.
                Currently supports numpy dense arrays, recursive lists ending with list or numpy.
            max_dim: Max number of dimensions the wrapped object has.
                If set to "auto", max dimensions will be inferred. This comes at compute cost.
        """
        self.o = o
        self.max_dim = max_dim
        if self.max_dim == "auto":
            self.max_dim = UnifiedDataHandler.max_dim(o)

    def __repr__(self) -> AnyStr:
        """ Override default repr for human readability.

        Returns:
            String to display.
        """
        return f"{self.__class__.__name__}({self.o.__repr__()})"

    def __getitem__(self, item: Any) -> Any:
        """ Consistent slicing into wrapped object.

        Args:
            item: Slicing key of type integer or slice.

        Returns:
            Sliced object.

        Raises:
            ValueError: If slicing is not compatible with wrapped object.
        """
        # Turn item into tuple if not already.
        index_tup = unify_slice(item, self.max_dim)

        # Slice according to object type.
        return UnifiedDataHandler.slice(self.o, index_tup, self.max_dim)


def unify_slice(item: Any, max_dim: int, alias_lookup=None) -> Tuple:
    """ Resolves aliases and ellipses in a slice item.

    Args:
        item: Slicing key that is passed to __getitem__.
        max_dim: Max dimension of object to be sliced.
        alias_lookup: AliasLookup structure.

    Returns:
        A tuple representation of the item.
    """
    item = _normalize_slice_key(item)
    index_tup = _normalize_subkey_types(item)
    index_tup = _handle_newaxis_ellipses(index_tup, max_dim)
    if alias_lookup:
        index_tup = _handle_aliases(index_tup, alias_lookup)
    return index_tup


def _normalize_subkey_types(index_tup: Tuple) -> Tuple:
    """ Casts subkeys into basic types such as int.

    Args:
        key: Slicing key that is passed within __getitem__.

    Returns:
        Tuple with subkeys casted to basic types.
    """
    new_index_tup = []  # Gets casted to tuple at the end

    np_int_types = {
        "int8",
        "int16",
        "int32",
        "int64",
        "uint8",
        "uint16",
        "uint32",
        "uint64",
    }
    for subkey in index_tup:
        if _safe_isinstance(subkey, "numpy", np_int_types):
            new_subkey = int(subkey)
        elif _safe_isinstance(subkey, "numpy", "ndarray"):
            if len(subkey.shape) == 1:
                new_subkey = subkey.tolist()
            else:
                raise ValueError(f"Cannot use array of shape {subkey.shape} as subkey.")
        else:
            new_subkey = subkey

        new_index_tup.append(new_subkey)
    return tuple(new_index_tup)


def _normalize_slice_key(key: Any) -> Tuple:
    """ Normalizes slice key into always being a top-level tuple.

    Args:
        key: Slicing key that is passed within __getitem__.

    Returns:
        Expanded slice as a tuple.
    """
    if not isinstance(key, tuple):
        return (key,)
    else:
        return key


def _handle_newaxis_ellipses(index_tup: Tuple, max_dim: int) -> Tuple:
    """ Expands newaxis and ellipses within a slice for simplification.
    This code is mostly adapted from: https://github.com/clbarnes/h5py_like/blob/master/h5py_like/shape_utils.py#L111

    Args:
        index_tup: Slicing key as a tuple.
        max_dim: Maximum number of dimensions in the respective sliceable object.

    Returns:
        Expanded slice as a tuple.
    """
    non_indexes = (None, Ellipsis)
    concrete_indices = sum(idx not in non_indexes for idx in index_tup)
    index_list = []
    # newaxis_at = []
    has_ellipsis = False
    int_count = 0
    for item in index_tup:
        if isinstance(item, numbers.Number):
            int_count += 1

        # NOTE: If we need locations of new axis, re-enable this.
        if item is None:  # pragma: no cover
            pass
            # newaxis_at.append(len(index_list) + len(newaxis_at) - int_count)
        elif item == Ellipsis:
            if has_ellipsis:  # pragma: no cover
                raise IndexError("an index can only have a single ellipsis ('...')")
            has_ellipsis = True
            initial_len = len(index_list)
            while len(index_list) + (concrete_indices - initial_len) < max_dim:
                index_list.append(slice(None))
        else:
            index_list.append(item)

    if len(index_list) > max_dim:  # pragma: no cover
        raise IndexError("too many indices for array")
    while len(index_list) < max_dim:
        index_list.append(slice(None))

    # return index_list, newaxis_at
    return tuple(index_list)


def _handle_aliases(index_tup: Tuple, alias_lookup) -> Tuple:
    new_index_tup = []

    def resolve(item, dim):
        if isinstance(item, slice):
            return item
        # Replace element if in alias lookup, otherwise use original.
        item = alias_lookup.get(dim, item, item)
        return item

    # Go through each element within the index and resolve if needed.
    for dim, item in enumerate(index_tup):
        if isinstance(item, list):
            new_item = []
            for sub_item in item:
                new_item.append(resolve(sub_item, dim))
        else:
            new_item = resolve(item, dim)
        new_index_tup.append(new_item)

    return tuple(new_index_tup)


class Tracked(AtomicSlicer):
    """ Tracked defines an object that slicer wraps."""

    def __init__(self, o: Any, dim: Union[int, List, tuple, None, str] = "auto"):
        """ Defines an object that will be wrapped by slicer.

        Args:
            o: Object that will be tracked for slicer.
            dim: Target dimension(s) slicer will index on for this object.
        """
        super().__init__(o)

        # Protected attribute that can be overriden.
        self._name = None

        # Place dim into coordinate form.
        if dim == "auto":
            self.dim = list(range(self.max_dim))
        elif dim is None:
            self.dim = []
        elif isinstance(dim, int):
            self.dim = [dim]
        elif isinstance(dim, list):
            self.dim = dim
        elif isinstance(dim, tuple):
            self.dim = list(dim)
        else:  # pragma: no cover
            raise ValueError(f"Cannot handle dim of type: {type(dim)}")


class Obj(Tracked):
    """ An object that slicer wraps. """
    def __init__(self, o, dim="auto"):
        super().__init__(o, dim)


class Alias(Tracked):
    """ Defines a tracked object as well as additional __getitem__ keys. """
    def __init__(self, o, dim):
        if not (
            isinstance(dim, int) or (isinstance(dim, (list, tuple)) and len(dim) <= 1)
        ):  # pragma: no cover
            raise ValueError("Aliases must track a single dimension")
        super().__init__(o, dim)


class AliasLookup:
    def __init__(self, aliases):
        self._lookup = {}

        # Populate lookup and merge indexes.
        for _, alias in aliases.items():
            self.update(alias)

    def update(self, alias):
        if alias.dim is None or len(alias.dim) == 0:
            return

        dim = alias.dim[0]
        if dim not in self._lookup:
            self._lookup[dim] = {}

        dim_lookup = self._lookup[dim]
        # NOTE: Alias must be backed by either a list or dictionary.
        itr = enumerate(alias.o) if isinstance(alias.o, list) else alias.o.items()
        for i, x in itr:
            if x not in dim_lookup:
                dim_lookup[x] = set()
            dim_lookup[x].add(i)

    def delete(self, alias):
        '''Delete an alias that exists from lookup'''
        dim = alias.dim[0]
        dim_lookup = self._lookup[dim]
        # NOTE: Alias must be backed by either a list or dictionary.
        itr = enumerate(alias.o) if isinstance(alias.o, list) else alias.o.items()
        for i, x in itr:
            del dim_lookup[x]

    def get(self, dim, target, default=None):
        if dim not in self._lookup:
            return default

        indexes = self._lookup[dim].get(target, None)
        if indexes is None:
            return default

        if len(indexes) == 1:
            return next(iter(indexes))
        else:
            return list(indexes)


def resolve_dim(slicer_index: Tuple, slicer_dim: List) -> List:
    """ Extracts new dim after applying slicing index and maps it back to the original index list. """

    new_slicer_dim = []
    reduced_mask = []

    for _, curr_idx in enumerate(slicer_index):
        if isinstance(curr_idx, (tuple, list, slice)):
            reduced_mask.append(0)
        else:
            reduced_mask.append(1)

    for curr_dim in slicer_dim:
        if reduced_mask[curr_dim] == 0:
            new_slicer_dim.append(curr_dim - sum(reduced_mask[:curr_dim]))

    return new_slicer_dim


def reduced_o(tracked: Tracked) -> Union[List, Any]:
    os = [t.o for t in tracked]
    os = os[0] if len(os) == 1 else os
    return os


class BaseHandler:
    @classmethod
    @abstractmethod
    def head_slice(cls, o, index_tup, max_dim):
        raise NotImplementedError()  # pragma: no cover

    @classmethod
    @abstractmethod
    def tail_slice(cls, o, tail_index, max_dim, flatten=True):
        raise NotImplementedError()  # pragma: no cover

    @classmethod
    @abstractmethod
    def max_dim(cls, o):
        raise NotImplementedError()  # pragma: no cover

    @classmethod
    def default_alias(cls, o):
        return []


class SeriesHandler(BaseHandler):
    @classmethod
    def head_slice(cls, o, index_tup, max_dim):
        head_index = index_tup[0]
        is_element = True if isinstance(head_index, int) else False
        sliced_o = o.iloc[head_index]

        return is_element, sliced_o, 1

    @classmethod
    def tail_slice(cls, o, tail_index, max_dim, flatten=True):
        # NOTE: Series only has one dimension,
        #       call slicer again to end the recursion.
        return AtomicSlicer(o, max_dim=max_dim)[tail_index]

    @classmethod
    def max_dim(cls, o):
        return len(o.shape)

    @classmethod
    def default_alias(cls, o):
        index_alias = Alias(o.index.to_list(), 0)
        index_alias._name = "index"
        return [index_alias]


class DataFrameHandler(BaseHandler):
    @classmethod
    def head_slice(cls, o, index_tup, max_dim):
        # NOTE: At head slice, we know there are two fixed dimensions.
        cut_index = index_tup
        is_element = True if isinstance(cut_index[-1], int) else False
        sliced_o = o.iloc[cut_index]

        return is_element, sliced_o, 2

    @classmethod
    def tail_slice(cls, o, tail_index, max_dim, flatten=True):
        # NOTE: Dataframe has fixed dimensions,
        #       call slicer again to end the recursion.
        return AtomicSlicer(o, max_dim=max_dim)[tail_index]

    @classmethod
    def max_dim(cls, o):
        return len(o.shape)

    @classmethod
    def default_alias(cls, o):
        index_alias = Alias(o.index.to_list(), 0)
        index_alias._name = "index"
        column_alias = Alias(o.columns.to_list(), 1)
        column_alias._name = "columns"
        return [index_alias, column_alias]


class ArrayHandler(BaseHandler):
    @classmethod
    def head_slice(cls, o, index_tup, max_dim):
        # Check if head is string
        head_index, tail_index = index_tup[0], index_tup[1:]
        cut = 1

        for sub_index in tail_index:
            if isinstance(sub_index, str) or cut == len(o.shape):
                break
            cut += 1

        # Process native array dimensions
        cut_index = index_tup[:cut]
        is_element = any([True if isinstance(x, int) else False for x in cut_index])
        sliced_o = o[cut_index]

        return is_element, sliced_o, cut

    @classmethod
    def tail_slice(cls, o, tail_index, max_dim, flatten=True):
        if flatten:
            # NOTE: If we're dealing with a scipy matrix,
            #       we have to manually flatten it ourselves
            #       to keep consistent to the rest of slicer's API.
            if _safe_isinstance(o, "scipy.sparse.csc", "csc_matrix"):
                return AtomicSlicer(o.toarray().flatten(), max_dim=max_dim)[tail_index]
            elif _safe_isinstance(o, "scipy.sparse.csr", "csr_matrix"):
                return AtomicSlicer(o.toarray().flatten(), max_dim=max_dim)[tail_index]
            elif _safe_isinstance(o, "scipy.sparse.dok", "dok_matrix"):
                return AtomicSlicer(o.toarray().flatten(), max_dim=max_dim)[tail_index]
            elif _safe_isinstance(o, "scipy.sparse.lil", "lil_matrix"):
                return AtomicSlicer(o.toarray().flatten(), max_dim=max_dim)[tail_index]
            else:
                return AtomicSlicer(o, max_dim=max_dim)[tail_index]
        else:
            inner = [AtomicSlicer(e, max_dim=max_dim)[tail_index] for e in o]
            if _safe_isinstance(o, "numpy", "ndarray"):
                import numpy
                if len(inner) > 0 and hasattr(inner[0], "__len__"):
                    ragged = not all(len(x) == len(inner[0]) for x in inner)
                else:
                    ragged = False
                if ragged:
                    return numpy.array(inner, dtype=object)
                else:
                    return numpy.array(inner)
            elif _safe_isinstance(o, "torch", "Tensor"):
                import torch

                if len(inner) > 0 and isinstance(inner[0], torch.Tensor):
                    return torch.stack(inner)
                else:
                    return torch.tensor(inner)
            elif _safe_isinstance(o, "scipy.sparse.csc", "csc_matrix"):
                from scipy.sparse import vstack
                out = vstack(inner, format='csc')
                return out
            elif _safe_isinstance(o, "scipy.sparse.csr", "csr_matrix"):
                from scipy.sparse import vstack
                out = vstack(inner, format='csr')
                return out
            elif _safe_isinstance(o, "scipy.sparse.dok", "dok_matrix"):
                from scipy.sparse import vstack
                out = vstack(inner, format='dok')
                return out
            elif _safe_isinstance(o, "scipy.sparse.lil", "lil_matrix"):
                from scipy.sparse import vstack
                out = vstack(inner, format='lil')
                return out
            else:
                raise ValueError(f"Cannot handle type {type(o)}.")  # pragma: no cover

    @classmethod
    def max_dim(cls, o):
        if _safe_isinstance(o, "numpy", "ndarray") and o.dtype == "object":
            return max([UnifiedDataHandler.max_dim(x) for x in o], default=-1) + 1
        else:
            return len(o.shape)


class DictHandler(BaseHandler):
    @classmethod
    def head_slice(cls, o, index_tup, max_dim):
        head_index = index_tup[0]
        if isinstance(head_index, (tuple, list)) and len(index_tup) == 0:
            return False, o, 1

        if isinstance(head_index, (list, tuple)):
            return (
                False,
                {
                    sub_index: AtomicSlicer(o, max_dim=max_dim)[sub_index]
                    for sub_index in head_index
                },
                1,
            )
        elif isinstance(head_index, slice):
            if head_index == slice(None, None, None):
                return False, o, 1
            return False, o[head_index], 1
        else:
            return True, o[head_index], 1

    @classmethod
    def tail_slice(cls, o, tail_index, max_dim, flatten=True):
        if flatten:
            return AtomicSlicer(o, max_dim=max_dim)[tail_index]
        else:
            return {
                k: AtomicSlicer(e, max_dim=max_dim)[tail_index] for k, e in o.items()
            }

    @classmethod
    def max_dim(cls, o):
        return max([UnifiedDataHandler.max_dim(x) for x in o.values()], default=-1) + 1


class ListTupleHandler(BaseHandler):
    @classmethod
    def head_slice(cls, o, index_tup, max_dim):
        head_index = index_tup[0]
        if isinstance(head_index, (tuple, list)) and len(index_tup) == 0:
            return False, o, 1

        if isinstance(head_index, (list, tuple)):
            if len(head_index) == 0:
                return False, o, 1
            else:
                results = [
                    AtomicSlicer(o, max_dim=max_dim)[sub_index]
                    for sub_index in head_index
                ]
                results = tuple(results) if isinstance(o, tuple) else results
                return False, results, 1
        elif isinstance(head_index, slice):
            return False, o[head_index], 1
        elif isinstance(head_index, int):
            return True, o[head_index], 1
        else:  # pragma: no cover
            raise ValueError(f"Invalid key {head_index} for {o}")

    @classmethod
    def tail_slice(cls, o, tail_index, max_dim, flatten=True):
        if flatten:
            return AtomicSlicer(o, max_dim=max_dim)[tail_index]
        else:
            results = [AtomicSlicer(e, max_dim=max_dim)[tail_index] for e in o]
            return tuple(results) if isinstance(o, tuple) else results

    @classmethod
    def max_dim(cls, o):
        return max([UnifiedDataHandler.max_dim(x) for x in o], default=-1) + 1


class UnifiedDataHandler:
    """ Registry that maps types to their unified slice calls."""

    """ Class attribute that maps type to their unified slice calls."""
    type_map = {
        ("builtins", "list"): ListTupleHandler,
        ("builtins", "tuple"): ListTupleHandler,
        ("builtins", "dict"): DictHandler,
        ("torch", "Tensor"): ArrayHandler,
        ("numpy", "ndarray"): ArrayHandler,
        ("scipy.sparse.csc", "csc_matrix"): ArrayHandler,
        ("scipy.sparse.csr", "csr_matrix"): ArrayHandler,
        ("scipy.sparse.dok", "dok_matrix"): ArrayHandler,
        ("scipy.sparse.lil", "lil_matrix"): ArrayHandler,
        ("pandas.core.frame", "DataFrame"): DataFrameHandler,
        ("pandas.core.series", "Series"): SeriesHandler,
    }

    @classmethod
    def slice(cls, o, index_tup, max_dim):
        # NOTE: Unified handles base cases such as empty tuples, which
        #       specialized handlers do not.
        if isinstance(index_tup, (tuple, list)) and len(index_tup) == 0:
            return o

        # Slice as delegated by data handler.
        o_type = _type_name(o)
        head_slice = cls.type_map[o_type].head_slice
        tail_slice = cls.type_map[o_type].tail_slice

        is_element, sliced_o, cut = head_slice(o, index_tup, max_dim)
        out = tail_slice(sliced_o, index_tup[cut:], max_dim - cut, is_element)
        return out

    @classmethod
    def max_dim(cls, o):
        o_type = _type_name(o)
        if o_type not in cls.type_map:
            return 0
        return cls.type_map[o_type].max_dim(o)

    @classmethod
    def default_alias(cls, o):
        o_type = _type_name(o)
        if o_type not in cls.type_map:
            return {}
        return cls.type_map[o_type].default_alias(o)


def _type_name(o: object) -> Tuple[str, str]:
    return _handle_module_aliases(o.__class__.__module__), o.__class__.__name__


def _safe_isinstance(
    o: object, module_name: str, type_name: Union[str, set, tuple]
) -> bool:
    o_module, o_type = _type_name(o)
    if isinstance(type_name, str):
        return o_module == module_name and o_type == type_name
    else:
        return o_module == module_name and o_type in type_name


def _handle_module_aliases(module_name):
    # scipy modules such as "scipy.sparse.csc" were renamed to "scipy.sparse._csc" in v1.8
    # Standardise by removing underscores for compatibility with either name
    # Else just pass module name unchanged
    module_map = {
        "scipy.sparse._csc": "scipy.sparse.csc",
        "scipy.sparse._csr": "scipy.sparse.csr",
        "scipy.sparse._dok": "scipy.sparse.dok",
        "scipy.sparse._lil": "scipy.sparse.lil",
    }
    return module_map.get(module_name, module_name)
